import logging

from datalad.distribution import siblings as mod_siblings
from datalad.support.annexrepo import AnnexRepo
from datalad.support.exceptions import (
    AccessDeniedError,
    AccessFailedError,
    CapturedException,
)

# use same logger as -core
lgr = logging.getLogger('datalad.distribution.siblings')

# This function is taken from datalad-core@2ed709613ecde8218a215dcb7d74b4a352825685
# datalad/distribution/siblings.py
# Changes
# - removed credential lookup for webdav-remotes
# - exception logging via CapturedException
def _enable_remote(ds, repo, name, res_kwargs, **unused_kwargs):
    result_props = dict(
        action='enable-sibling',
        path=ds.path,
        type='sibling',
        name=name,
        **res_kwargs)

    if not isinstance(repo, AnnexRepo):
        yield dict(
            result_props,
            status='impossible',
            message='cannot enable sibling of non-annex dataset')
        return

    if name is None:
        yield dict(
            result_props,
            status='error',
            message='require `name` of sibling to enable')
        return

    # get info on special remote
    sp_remotes = {v['name']: dict(v, uuid=k) for k, v in repo.get_special_remotes().items()}
    remote_info = sp_remotes.get(name, None)

    if remote_info is None:
        yield dict(
            result_props,
            status='impossible',
            message=("cannot enable sibling '%s', not known", name))
        return

    try:
        repo.enable_remote(name)
        result_props['status'] = 'ok'
    except (AccessDeniedError, AccessFailedError) as e:
        CapturedException(e)
        result_props['status'] = 'error'
        # TODO should use proper way of injecting exceptions in result records
        result_props['message'] = str(e)

    yield result_props


# apply patch
lgr.debug('Apply datalad-next patch to siblings.py:_enable_remote')
mod_siblings._enable_remote = _enable_remote



from urllib.parse import urlparse
import requests
import www_authenticate
from datalad.downloaders.http import DEFAULT_USER_AGENT 

__all__ = ['probe_url', 'get_auth_realm']


def probe_url(url, timeout=10.0, headers=None):
    """Probe a HTTP(S) URL for redirects and authentication needs

    This functions performs a HEAD request against the given URL,
    while waiting at most for the given timeout duration for
    a server response.

    Parameters
    ----------
    url: str
      URL to probe
    timeout: float, optional
      Maximum time to wait for a server response to the probe
    headers: dict, optional
      Any custom headers to use for the probe request. If none are
      provided, or the provided headers contain no 'user-agent'
      field, the default DataLad user agent is added automatically.

    Returns
    -------
    str or None, dict
      The first value is the URL against the final request was
      performed, after following any redirects and applying
      normalizations.

      The second value is a mapping with a particular set of
      properties inferred from probing the webserver. The following
      key-value pairs are supported:

      - 'is_redirect' (bool), True if any redirection occurred. This
        boolean property is a more accurate test than comparing
        input and output URL
      - 'status_code' (int), HTTP response code (of the final request
        in case of redirection).
      - 'auth' (dict), present if the final server response contained any
        'www-authenticate' headers, typically the case for 401 responses.
        The dict contains a mapping of server-reported authentication
        scheme names (e.g., 'basic', 'bearer') to their respective
        properties (dict). These can be any nature and number, depending
        on the respective authentication scheme. Most notably, they
        may contain a 'realm' property that can be used to determine
        suitable credentials for authentication.

    Raises
    ------
    requests.RequestException
      May raise any exception of the `requests` package, most notably
      `ConnectionError`, `Timeout`, `TooManyRedirects`, etc.
    """
    hdrs = {'user-agent': DEFAULT_USER_AGENT}
    if headers is None:
        headers = hdrs
    elif 'user-agent' not in headers:
        headers.update(hdrs)

    props = {}
    req = requests.head(
        url,
        allow_redirects=True,
        timeout=timeout,
        headers=headers,
    )
    if 'www-authenticate' in req.headers:
        props['auth'] = www_authenticate.parse(
            req.headers['www-authenticate'])
    props['is_redirect'] = True if req.history else False
    props['status_code'] = req.status_code
    return req.url, props


def get_auth_realm(url, auth_info, scheme=None):
    """Determine an authentication realm identifier from a HTTP response.

    Examples
    --------
    Robustly determine a realm identifier for any URL::

       >>> url, props = probe_url('https://fz-juelich.sciebo.de/...')
       >>> get_auth_realm(url, props.get('auth'))
       'https://fz-juelich.sciebo.de/login'

    Parameters
    ----------
    url: str
      A URL as returned by `probe_url()`
    auth_info: dict
      A mapping of supported authentication schemes to the properties
      (i.e., a 'www-authenticate' response header), as returned by
      `probe_url()`'s 'auth' property.
    scheme: str, optional
      Which specific authentication to report a realm for, in case
      multiple are supported (such as 'basic', or 'token').
      If not given, the first (if any) reported authentication
      scheme is reported on.

    Returns
    -------
    str
      A server-specific realm identifier
    """
    if not auth_info:
        # no info from the server on what it needs
        # out best bet is the URL itself
        return url
    if scheme:
        auth_info = auth_info[scheme]
    else:
        scheme, auth_info = auth_info.popitem()
    # take any, but be satisfied with none too
    realm = auth_info.get('realm', '')
    # a realm is supposed to indicate a validity scope of a credential
    # on a server. so we make sure to have the return realm identifier
    # actually indicate a server too, in order to make it suitable for
    # a global credential lookup
    if _is_valid_url(realm):
        # the realm is already a valid URL with a server specification.
        # we can simply relay it as such, following the admins' judgement
        return realm
    else:
        # the realm was just some kind of string. we prefix it with the
        # netloc of the given URL (ignoring its path) to achieve
        # the same server-specific realm semantics
        parsed = urlparse(url)
        return '{scheme}://{netloc}{slash}{realm}'.format(
            scheme=parsed.scheme,
            netloc=parsed.netloc,
            slash='' if realm.startswith('/') else'/',
            realm=realm,
        )



def _is_valid_url(url):
    try:
        parsed = urlparse(url)
        return all([parsed.scheme, parsed.netloc])
    except:
        return False



import logging
from pathlib import Path
from typing import (
    Dict,
    Generator,
    Iterable,
    Optional,
    Union,
)
from unittest.mock import patch

import datalad.core.distributed.push as push
from datalad.distribution.dataset import Dataset
from datalad.runner.exception import CommandError
from datalad.support.annexrepo import AnnexRepo
from datalad.support.constraints import EnsureChoice
from datalad.support.exceptions import CapturedException
from datalad.support.param import Parameter
from datalad_next.credman import CredentialManager
from datalad_next.utils import (
    get_specialremote_credential_envpatch,
    get_specialremote_credential_properties,
    needs_specialremote_credential_envpatch,
)


lgr = logging.getLogger('datalad.core.distributed.push')


def _is_export_remote(remote_info: Optional[Dict]) -> bool:
    """Check if remote_info is valid and has exporttree set to "yes"

    Parameters
    ----------
    remote_info: Optional[Dict]
        Optional dictionary the contains git annex special.

    Returns
    -------
    bool
        True if exporttree key is contained in remote_info and is set to yes,
        else False.
    """
    if remote_info is not None:
        return remote_info.get("exporttree") == "yes"
    return False


def _get_credentials(ds: Dataset,
                     remote_info: Dict
                     ) -> Optional[Dict]:

    # Check for credentials
    params = {
        "type": remote_info.get("type"),
        "url": remote_info.get("url")
    }
    credentials = None
    credential_properties = get_specialremote_credential_properties(params)
    if credential_properties:
        # TODO: lower prio: factor this if clause out, also used in
        #  create_sibling_webdav.py
        credential_manager = CredentialManager(ds.config)
        credentials = (credential_manager.query(
            _sortby='last-used',
            **credential_properties) or [(None, None)])[0][1]
    return credentials


def get_export_records(repo: AnnexRepo) -> Generator:
    """Read exports that git-annex recorded in its 'export.log'-file

    Interpret the lines in export.log. Each line has the following structure:

        time-stamp " " source-annex-uuid ":" destination-annex-uuid " " treeish

    Parameters
    ----------
    repo: AnnexRepo
        The annex repo from which exports should be determined

    Returns
    -------
    Generator
        Generator yielding one dictionary for each export entry in git-annex.
        Each dictionary contains the keys: "timestamp", "source-annex-uuid",
        "destination-annex-uuid", "treeish". The timestamp-value is a float,
        all other values are strings.
    """
    try:
        for line in repo.call_git_items_(["cat-file", "blob", "git-annex:export.log"]):
            result_dict = dict(zip(
                [
                    "timestamp",
                    "source-annex-uuid",
                    "destination-annex-uuid",
                    "treeish"
                ],
                line.replace(":", " ").split()
            ))
            result_dict["timestamp"] = float(result_dict["timestamp"][:-1])
            yield result_dict
    except CommandError as command_error:
        # Some errors indicate that there was no export yet.
        expected_errors = (
            "fatal: Not a valid object name git-annex:export.log",
        )
        if command_error.stderr.strip() in expected_errors:
            return
        raise


def _get_export_log_entry(repo: AnnexRepo,
                          target_uuid: str
                          ) -> Optional[Dict]:
    target_entries = [
        entry
        for entry in repo.get_export_records()
        if entry["destination-annex-uuid"] == target_uuid]

    if not target_entries:
        return None
    return sorted(target_entries, key=lambda e: e["timestamp"])[-1]


def _is_valid_treeish(repo: AnnexRepo,
                      export_entry: Dict,
                      ) -> bool:

    # Due to issue https://github.com/datalad/datalad-next/issues/39
    # fast-forward validation has to be re-designed.
    return True
    #for line in repo.call_git_items_(["log", "--pretty=%H %T"]):
    #    commit_hash, treeish = line.split()
    #    if treeish == export_entry["treeish"]:
    #        return True
    #return False


def _transfer_data(repo: AnnexRepo,
                   ds: Dataset,
                   target: str,
                   content: Iterable,
                   data: str,
                   force: Optional[str],
                   jobs: Optional[Union[str, int]],
                   res_kwargs: Dict,
                   got_path_arg: bool
                   ) -> Generator:

    target_uuid, remote_info = ([
        (uuid, info) for uuid, info in repo.get_special_remotes().items()
        if info.get("name") == target] or [(None, None)])[0]

    if not _is_export_remote(remote_info):
        yield from push._push_data(
            ds,
            target,
            content,
            data,
            force,
            jobs,
            res_kwargs.copy(),
            got_path_arg=got_path_arg,
        )
        return

    # TODO:
    #  - check for configuration entries, e.g. what to export

    lgr.debug(f"Exporting HEAD of {ds} to remote {remote_info}")

    if ds.config.getbool('remote.{}'.format(target), 'annex-ignore', False):
        lgr.debug(
            "Target '%s' is set to annex-ignore, exclude from data-export.",
            target)
        return

    if force not in ("all", "export"):
        export_entry = _get_export_log_entry(repo, target_uuid)
        if export_entry:
            if export_entry["source-annex-uuid"] != repo.uuid:
                yield dict(
                    **res_kwargs,
                    status="error",
                    message=f"refuse to export to {target}, because the "
                            f"last known export came from another repo "
                            f"({export_entry['source-annex-uuid']}). Use "
                            f"--force=export to enforce the export anyway.")
                return
            if not _is_valid_treeish(repo, export_entry):
                yield dict(
                    **res_kwargs,
                    status="error",
                    message=f"refuse to export to {target}, because the "
                            f"current state is not a fast-forward of the "
                            f"last known exported state. Use "
                            f"--force=export to enforce the export anyway.")
                return

    credentials = _get_credentials(ds, remote_info)

    # If we have credentials, check whether we require an environment patch
    env_patch = {}
    remote_type = remote_info.get("type")
    if credentials and needs_specialremote_credential_envpatch(remote_type):
        env_patch = get_specialremote_credential_envpatch(
            remote_type,
            credentials)

    res_kwargs['target'] = target

    with patch.dict('os.environ', env_patch):
        try:
            for result in repo._call_annex_records_items_(
                [
                    "export", "HEAD",
                    "--to", target
                ],
                progress=True
            ):
                yield {
                    **res_kwargs,
                    "action": "copy",
                    "status": "ok",
                    "path": str(Path(res_kwargs["path"]) / result["file"])
                }

        except CommandError as cmd_error:
            ce = CapturedException(cmd_error)
            yield {
                **res_kwargs,
                "action": "copy",
                "status": "error",
                "message": str(ce),
                "exception": ce
            }


lgr.debug("Patching datalad.core.distributed.push._transfer_data")
push._transfer_data = _transfer_data


lgr.debug("Patching datalad.core.distributed.push.Push docstring and parameters")
push.Push.__doc__ += """\


    The following feature is added by the datalad-next extension:

    If a target is a git-annex special remote that has "exporttree" set to
    "yes", push will call 'git-annex export' to export the current HEAD to the
    remote target. This will usually result in a copy of the file tree, to which
    HEAD refers, on the remote target. A git-annex special remote with
    "exporttree" set to "yes" can, for example, be created with the datalad
    command "create-sibling-webdav" with the option "--mode=filetree" or 
    "--mode=filetree-only".
"""
push.Push._params_["force"] = Parameter(
    args=("-f", "--force",),
    doc="""force particular operations, possibly overruling safety
    protections or optimizations: use --force with git-push ('gitpush');
    do not use --fast with git-annex copy ('checkdatapresent'); force an
    annex export (to git annex remotes with "exporttree" set to "yes");
    combine all force modes ('all').""",
    constraints=EnsureChoice(
        'all', 'gitpush', 'checkdatapresent', 'export', None))


lgr.debug("Patching datalad.support.AnnexRepo.get_export_records (new method)")
AnnexRepo.get_export_records = get_export_records



import logging
import os
import re

from datalad.support.annexrepo import AnnexRepo
from datalad.support.exceptions import (
    CommandError,
)
from datalad.utils import ensure_list

from datalad_next.credman import CredentialManager
from datalad_next.utils import (
    get_specialremote_credential_envpatch,
    get_specialremote_param_dict,
    get_specialremote_credential_properties,
    needs_specialremote_credential_envpatch,
)


# reuse logger from -core, despite the unconventional name
lgr = logging.getLogger('datalad.annex')


# This function is taken from datalad-core@2ed709613ecde8218a215dcb7d74b4a352825685
# datalad/support/annexrepo.py:AnnexRepo
# Changes
# - raise exceptions carry context of the original command error
def annexRepo__enable_remote(self, name, options=None, env=None):
    """Enables use of an existing special remote

    Parameters
    ----------
    name: str
        name, the special remote was created with
    options: list, optional
    """
    # MIH thinks there should be no `env` argument at all
    # https://github.com/datalad/datalad/issues/5162
    # if it would not be there, this whole dance is pretty much
    # obsolete
    env = env or self._git_runner.env

    # an enableremote can do pretty much anything, including a type change.
    # in order to be able to determine whether credentials *will* be needed,
    # we have to look ahead and form the special remote parameters that will
    # be there at the end -- more or less

    # pull info for present config
    sp_remotes = {v['name']: dict(v, uuid=k) for k, v in self.get_special_remotes().items()}
    remote_info = sp_remotes.get(name, {})
    # TODO if remote_info is empty, we can fail right here
    if options:
        # and now update with given params
        remote_info.update(get_specialremote_param_dict(options))

    # careful here, `siblings()` also calls this for regular remotes, check
    # for a known type
    if 'type' in remote_info \
            and needs_specialremote_credential_envpatch(remote_info['type']):
        # see if we can identify any matching credentials
        credprops = get_specialremote_credential_properties(remote_info)
        credman = None
        credspec = None
        if credprops:
            credman = CredentialManager(self.config)
            creds = credman.query(_sortby='last-used', **credprops)
            if creds:
                # found one
                credspec = creds[0]
        # TODO manual entry could be supported here too! (also see at the end)
        if env:
            env.copy()

        if credspec:
            credpatch = get_specialremote_credential_envpatch(
                remote_info['type'], credspec[1])
            if credpatch:
                if not env:
                    env = os.environ.copy()
                env.update(credpatch)

    try:
        from unittest.mock import patch
        with patch.object(self._git_runner, 'env', env):
            # TODO: outputs are nohow used/displayed. Eventually convert to
            # to a generator style yielding our "dict records"
            self.call_annex(['enableremote', name] + ensure_list(options))
    except CommandError as e:
        if re.match(r'.*StatusCodeException.*statusCode = 401', e.stderr):
            raise AccessDeniedError(e.stderr) from e
        elif 'FailedConnectionException' in e.stderr:
            raise AccessFailedError(e.stderr) from e
        else:
            raise e
    self.config.reload()
    # TODO when manual credential entry is supported,
    # implement store-after-success here


lgr.debug('Apply datalad-next patch to annexrepo.py:AnnexRepo.enable_remote')
AnnexRepo.enable_remote = annexRepo__enable_remote
